/*
 *	Berkshire PCI-PC Watchdog Card Driver
 *
 *	Copyright (c) 1998-2002 Holger Eiboeck  <holger@eiboeck.de>,
 *				  All Rights Reserved.
 *
 *	http://www.pcwd.de/
 *
 *	(c) Copyright 2005 Wim Van Sebroeck <wim@iguana.be>.
 *
 *	Idea and parts of the code by Kenji Hollis <khollis@bitgate.com>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

/*
 *	Includes, defines, variables, module parameters, ...
 */

#include <linux/version.h>
#include <linux/config.h>	/* For CONFIG_WATCHDOG_NOWAYOUT/... */
#include <linux/module.h>	/* For module specific items */
#include <linux/moduleparam.h>	/* For new moduleparam's */
#include <linux/types.h>	/* For standard types (like size_t) */
#include <linux/errno.h>	/* For the -ENODEV/... values */
#include <linux/kernel.h>	/* For printk/panic/... */
#include <linux/delay.h>	/* For mdelay function */
#include <linux/miscdevice.h>	/* For MODULE_ALIAS_MISCDEV(WATCHDOG_MINOR) */
#include <linux/watchdog.h>	/* For the watchdog specific items */
#include <linux/notifier.h>	/* For notifier support */
#include <linux/reboot.h>	/* For reboot_notifier stuff */
#include <linux/init.h>		/* For __init/__exit/... */
#include <linux/fs.h>		/* For file operations */
#include <linux/pci.h>		/* For pci functions */
#include <linux/ioport.h>	/* For io-port access */
#include <linux/spinlock.h>	/* For spin_lock/spin_unlock/... */
#include <linux/string.h>	/* For string manipulations */
#ifdef CONFIG_PROC_FS
#include <linux/stat.h>		/* For S_IFREG/S_IRUGO/S_IWUSR */
#include <linux/proc_fs.h>	/* For create_proc_entry/remove_proc_entry/ ... */
#endif

#include <asm/uaccess.h>	/* For copy_to_user/put_user/... */
#include <asm/io.h>		/* For inb/outb/... */

/* Module and version information */
#define WATCHDOG_DRIVER_NAME "PCI-PC Watchdog"
#define WATCHDOG_NAME "pcwd_pci"
#define PFX WATCHDOG_NAME ": "

static char rcsid[] = "$Revision: 1.50 $$Date: 2005/10/09 20:21:00 $";

/* Stuff for the PCI ID's  */
#ifndef PCI_VENDOR_ID_QUICKLOGIC
#define PCI_VENDOR_ID_QUICKLOGIC    0x11e3
#endif

#ifndef PCI_DEVICE_ID_WATCHDOG_PCIPCWD
#define PCI_DEVICE_ID_WATCHDOG_PCIPCWD 0x5030
#endif

/*
 * These are the defines that describe the control status bits for the
 * PCI-PC Watchdog card.
 */
/* Port 1 : Control Status #1 */
#define WD_PCI_WTRP		0x01	/* Watchdog Trip status */
#define WD_PCI_HRBT		0x02	/* Watchdog Heartbeat */
#define WD_PCI_TTRP		0x04	/* Temperature Trip status */
#define WD_PCI_RL2A		0x08	/* Relay 2 Active */
#define WD_PCI_RL1A		0x10	/* Relay 1 Active */
#define WD_PCI_R2DS		0x40	/* Relay 2 Disable Temperature-trip/reset */
#define WD_PCI_RLY2		0x80	/* Activate Relay 2 on the board */
/* Port 2 : Control Status #2 */
#define WD_PCI_WDIS		0x10	/* Watchdog Disable */
#define WD_PCI_ENTP		0x20	/* Enable Temperature Trip Reset */
#define WD_PCI_WRSP		0x40	/* Watchdog wrote response */
#define WD_PCI_PCMD		0x80	/* PC has sent command */

/* according to documentation max. time to process a command for the pci
 * watchdog card is 100 ms, so we give it 150 ms to do it's job */
#define PCI_COMMAND_TIMEOUT	150

/* Watchdog's internal commands */
#define CMD_GET_STATUS				0x04
#define CMD_GET_FIRMWARE_VERSION		0x08
#define CMD_READ_ARM_TIME			0x10
#define CMD_WRITE_ARM_TIME			0x11
#define CMD_READ_NV_ARM_TIME			0x14
#define CMD_WRITE_NV_ARM_TIME			0x15
#define CMD_READ_WATCHDOG_TIMEOUT		0x18
#define CMD_WRITE_WATCHDOG_TIMEOUT		0x19
#define CMD_READ_NV_WATCHDOG_TIMEOUT		0x1C
#define CMD_WRITE_NV_WATCHDOG_TIMEOUT		0x1D
#define CMD_READ_RETRIGGER_COUNT		0x20
#define CMD_READ_EXTERNAL_RETRIGGER_COUNT	0x22
#define CMD_READ_RELAY_CONTROL			0x50
#define CMD_WRITE_RELAY_CONTROL			0x51
#define CMD_READ_TEMPERATURE_OFFSET		0x58
#define CMD_WRITE_TEMPERATURE_OFFSET		0x59
#define CMD_READ_BUZZER_CONTROL			0x5C	/* Only from fw-ver 1.50 */
#define CMD_WRITE_BUZZER_CONTROL		0x5D	/* Only from fw-ver 1.50 */
#define CMD_READ_AD_INPUT			0x60	/* Only from fw-ver 1.32 */
#define CMD_RESET_PC				0x80
#define CMD_GET_CLEAR_RESET_COUNT		0x84
#define CMD_READ_RESET_RELAY_PULSE_WIDTH	0x88	/* Only from fw-ver 1.32 */
#define CMD_WRITE_RESET_RELAY_PULSE_WIDTH	0x89	/* Only from fw-ver 1.32 */
#define CMD_READ_USER_ID_DATA			0xC0	/* Only from fw-ver 1.32 */
#define CMD_WRITE_USER_ID_DATA			0xC1	/* Only from fw-ver 1.32 */

/* We can only use 1 card due to the /dev/watchdog restriction */
static int cards_found;

/* pcipcwd_private.card_status */
#define DISABLED		0
#define ENABLED			1

/* mode of operation for relay 2 */
#define TEMPERATURE_TRIP	0	/* activate relay 2 during a temperature trip */
#define NO_TEMPERATURE_TRIP	1	/* don't do it */

/* internal variables */
static char pcwd_driver_version[40];
static int temp_panic;
static unsigned long is_active;
static char expect_release;
static struct {				/* this is private data for each PCI-PC watchdog card */
	char fw_ver_str[20];		/* The cards firmware version */
	int supports_temp;		/* Wether or not the card has a temperature device */
	int boot_status;		/* The card's boot status */
	unsigned long io_addr;		/* The cards I/O address */
	int card_status;		/* The card's status */
	unsigned int ping_counter;	/* A counter to keep track how many times the watchdog was tickled */
	int temp_mode_relay2;		/* mode of operation for relay 2 */
	spinlock_t io_lock;		/* the lock for io operations */
#ifdef CONFIG_PROC_FS
	spinlock_t proc_lock;		/* the lock for /proc/ operations */
	struct proc_dir_entry *proc_pcwd;	/* the /proc/ entry */
#endif
} pcipcwd_private;

/* module parameters */
#define CELSIUS		0
#define FAHRENHEIT	1
static int proc_temp_mode = CELSIUS;
module_param(proc_temp_mode, int, 0);
MODULE_PARM_DESC(proc_temp_mode, "which temperature mode to use in /proc/ 0=Celsius, 1=Fahrenheit (default=0)");

#define QUIET	0	/* Default */
#define VERBOSE	1	/* Verbose */
#define DEBUG	2	/* print fancy stuff too */
static int debug = QUIET;
module_param(debug, int, 0);
MODULE_PARM_DESC(debug, "Debug level: 0=Quiet, 1=Verbose, 2=Debug (default=0)");

#define WATCHDOG_HEARTBEAT 2	/* 2 sec default heartbeat */
static int heartbeat = WATCHDOG_HEARTBEAT;
module_param(heartbeat, int, 0);
MODULE_PARM_DESC(heartbeat, "Watchdog heartbeat in seconds. (0<heartbeat<65536, default=" __MODULE_STRING(WATCHDOG_HEARTBEAT) ")");

#ifdef CONFIG_WATCHDOG_NOWAYOUT
static int nowayout = 1;
#else
static int nowayout = 0;
#endif

module_param(nowayout, int, 0);
MODULE_PARM_DESC(nowayout, "Watchdog cannot be stopped once started (default=CONFIG_WATCHDOG_NOWAYOUT)");

/*
 *	Internal functions
 */

static void pcipcwd_get_version(void)
{
	char *rcsvers, *rcsdate, *tmp;

	rcsvers = strchr(rcsid, ' ');
	rcsvers++;
	tmp = strchr(rcsvers, ' ');
	*tmp++ = '\0';
	rcsdate = strchr(tmp, ' ');
	rcsdate++;
	tmp = strrchr(rcsdate, ' ');
	*tmp = '\0';

	sprintf(pcwd_driver_version, "%s (%s)", rcsvers, rcsdate);
}

static int send_command(int cmd, int *msb, int *lsb)
{
	int got_response, count;

	if (debug >= DEBUG)
		printk(KERN_DEBUG PFX "sending following data cmd=0x%02x msb=0x%02x lsb=0x%02x\n",
		cmd, *msb, *lsb);

	spin_lock(&pcipcwd_private.io_lock);
	/* If a command requires data it should be written first.
	 * Data for commands with 8 bits of data should be written to port 4.
	 * Commands with 16 bits of data, should be written as LSB to port 4
	 * and MSB to port 5.
	 * After the required data has been written then write the command to
	 * port 6. */
	outb_p(*lsb, pcipcwd_private.io_addr + 4);
	outb_p(*msb, pcipcwd_private.io_addr + 5);
	outb_p(cmd, pcipcwd_private.io_addr + 6);

	/* wait till the pci card processed the command, signaled by
	 * the WRSP bit in port 2 and give it a max. timeout of
	 * PCI_COMMAND_TIMEOUT to process */
	got_response = inb_p(pcipcwd_private.io_addr + 2) & WD_PCI_WRSP;
	for (count = 0; (count < PCI_COMMAND_TIMEOUT) && (!got_response); count++) {
		mdelay(1);
		got_response = inb_p(pcipcwd_private.io_addr + 2) & WD_PCI_WRSP;
	}

	if (debug >= DEBUG) {
		if (got_response) {
			printk(KERN_DEBUG PFX "time to process command was: %d ms\n",
				count);
		} else {
			printk(KERN_DEBUG PFX "card did not respond on command!\n");
		}
	}

	if (got_response) {
		/* read back response */
		*lsb = inb_p(pcipcwd_private.io_addr + 4);
		*msb = inb_p(pcipcwd_private.io_addr + 5);

		/* clear WRSP bit */
		inb_p(pcipcwd_private.io_addr + 6);

		if (debug >= DEBUG)
			printk(KERN_DEBUG PFX "received following data for cmd=0x%02x: msb=0x%02x lsb=0x%02x\n",
				cmd, *msb, *lsb);
	}

	spin_unlock(&pcipcwd_private.io_lock);

	return got_response;
}

/* send a command to watchdog card and give a 16 bit return value */
static int send_command_ret16(int cmd, int msb, int lsb)
{
	int m, l, response;

	m = msb;
	l = lsb;
	response = 0;

	if (send_command(cmd, &m, &l))
		response = (m * 256) + l;

	return response;
}

/* send a command to watchdog card and give a 8 bit return value */
static int send_command_ret8(int cmd, int msb, int lsb)
{
	return (send_command_ret16(cmd, msb, lsb) & 0x00FF);
}

static inline void pcipcwd_check_temperature_support(void)
{
	if (inb_p(pcipcwd_private.io_addr) != 0xF0)
		pcipcwd_private.supports_temp = 1;
}

static int pcipcwd_get_option_switches(void)
{
	int option_switches;

	option_switches = inb_p(pcipcwd_private.io_addr + 3);
	return option_switches;
}

static void pcipcwd_show_card_info(void)
{
	int got_fw_rev, fw_rev_major, fw_rev_minor;
	int option_switches;

	got_fw_rev = send_command(CMD_GET_FIRMWARE_VERSION, &fw_rev_major, &fw_rev_minor);
	if (got_fw_rev) {
		sprintf(pcipcwd_private.fw_ver_str, "%u.%02u", fw_rev_major, fw_rev_minor);
	} else {
		sprintf(pcipcwd_private.fw_ver_str, "<card no answer>");
	}

	/* Get switch settings */
	option_switches = pcipcwd_get_option_switches();

	printk(KERN_INFO PFX "Found card at port 0x%04x (Firmware: %s) %s temp option\n",
		(int) pcipcwd_private.io_addr, pcipcwd_private.fw_ver_str,
		(pcipcwd_private.supports_temp ? "with" : "without"));

	printk(KERN_INFO PFX "Option switches (0x%02x): Temperature Reset Enable=%s, Power On Delay=%s\n",
		option_switches,
		((option_switches & 0x10) ? "ON" : "OFF"),
		((option_switches & 0x08) ? "ON" : "OFF"));

	if (pcipcwd_private.boot_status & WDIOF_CARDRESET)
		printk(KERN_INFO PFX "Previous reset was caused by the Watchdog card\n");

	if (pcipcwd_private.boot_status & WDIOF_OVERHEAT)
		printk(KERN_INFO PFX "Card sensed a CPU Overheat\n");

	if (pcipcwd_private.boot_status == 0)
		printk(KERN_INFO PFX "No previous trip detected - Cold boot or reset\n");
}

static int pcipcwd_start(void)
{
	int stat_reg;

	spin_lock(&pcipcwd_private.io_lock);
	outb_p(0x00, pcipcwd_private.io_addr + 3);
	udelay(1000);

	stat_reg = inb_p(pcipcwd_private.io_addr + 2);
	spin_unlock(&pcipcwd_private.io_lock);

	if (stat_reg & WD_PCI_WDIS) {
		printk(KERN_ERR PFX "Card timer not enabled\n");
		return -1;
	}

	pcipcwd_private.card_status = ENABLED;
	if (debug >= VERBOSE)
		printk(KERN_DEBUG PFX "Watchdog started\n");

	return 0;
}

static int pcipcwd_stop(void)
{
	int stat_reg;

	spin_lock(&pcipcwd_private.io_lock);
	outb_p(0xA5, pcipcwd_private.io_addr + 3);
	udelay(1000);

	outb_p(0xA5, pcipcwd_private.io_addr + 3);
	udelay(1000);

	stat_reg = inb_p(pcipcwd_private.io_addr + 2);
	spin_unlock(&pcipcwd_private.io_lock);

	if (!(stat_reg & WD_PCI_WDIS)) {
		printk(KERN_ERR PFX "Card did not acknowledge disable attempt\n");
		return -1;
	}

	pcipcwd_private.card_status = DISABLED;
	if (debug >= VERBOSE)
		printk(KERN_DEBUG PFX "Watchdog stopped\n");

	return 0;
}

static int pcipcwd_keepalive(void)
{
	/* Re-trigger watchdog by writing to port 0 */
	spin_lock(&pcipcwd_private.io_lock);
	outb_p(0x42, pcipcwd_private.io_addr);	/* send out any data */
	spin_unlock(&pcipcwd_private.io_lock);
	if (pcipcwd_private.card_status == ENABLED)
		pcipcwd_private.ping_counter++;

	if (debug >= DEBUG)
		printk(KERN_DEBUG PFX "Watchdog keepalive signal send\n");

	return 0;
}

static int pcipcwd_set_heartbeat(int t)
{
	int t_msb = t / 256;
	int t_lsb = t % 256;

	if ((t < 0x0001) || (t > 0xFFFF))
		return -EINVAL;

	/* Write new heartbeat to watchdog */
	send_command(CMD_WRITE_WATCHDOG_TIMEOUT, &t_msb, &t_lsb);

	heartbeat = t;
	if (debug >= VERBOSE)
		printk(KERN_DEBUG PFX "New heartbeat: %d\n",
		       heartbeat);

	return 0;
}

static int pcipcwd_get_status(int *status)
{
	int control_status;

	*status=0;
	control_status = inb_p(pcipcwd_private.io_addr + 1);
	if (control_status & WD_PCI_WTRP)
		*status |= WDIOF_CARDRESET;
	if (control_status & WD_PCI_TTRP) {
		*status |= WDIOF_OVERHEAT;
		if (temp_panic)
			panic(PFX "Temperature overheat trip!\n");
	}

	if (debug >= DEBUG)
		printk(KERN_DEBUG PFX "Control Status #1: 0x%02x\n",
		       control_status);

	return 0;
}

static int pcipcwd_clear_status(void)
{
	int control_status;
	int msb;
	int reset_counter;

	if (debug >= VERBOSE)
		printk(KERN_INFO PFX "clearing watchdog trip status & LED\n");

	control_status = inb_p(pcipcwd_private.io_addr + 1);

	if (debug >= DEBUG) {
		printk(KERN_DEBUG PFX "status was: 0x%02x\n", control_status);
		printk(KERN_DEBUG PFX "sending: 0x%02x\n",
		       (control_status & WD_PCI_R2DS) | WD_PCI_WTRP);
	}

	/* clear trip status & LED and keep mode of relay 2 */
	outb_p((control_status & WD_PCI_R2DS) | WD_PCI_WTRP, pcipcwd_private.io_addr + 1);

	/* clear reset counter */
	msb=0;
	reset_counter=0xff;
	send_command(CMD_GET_CLEAR_RESET_COUNT, &msb, &reset_counter);

	if (debug >= DEBUG) {
		printk(KERN_DEBUG PFX "reset count was: 0x%02x\n",
		       reset_counter);
	}

	return 0;
}

static int pcipcwd_get_temperature(int *temperature, int temp_mode)
{
	*temperature = 0;
	if (!pcipcwd_private.supports_temp)
		return -ENODEV;

	spin_lock(&pcipcwd_private.io_lock);
	*temperature = inb_p(pcipcwd_private.io_addr);
	spin_unlock(&pcipcwd_private.io_lock);

	/*
	 * Convert celsius to fahrenheit, this is normally
	 * the decided 'standard' for this return value.
	 */
	if (temp_mode == FAHRENHEIT)
		*temperature = (*temperature * 9 / 5) + 32;

	if (debug >= DEBUG) {
		printk(KERN_DEBUG PFX "temperature is: %d%s\n",
		       *temperature, (temp_mode == CELSIUS) ? "C\n" :"F\n");
	}

	return 0;
}

#ifdef CONFIG_PROC_FS
static int
pcipcwd_reset_pc(void)
{
	int rc;

	printk(KERN_INFO PFX "reseting PC!\n");

	/* reset with a 5 second delay -
	 * user might want to umount, call sync, ... :) */

	rc = send_command_ret8(CMD_RESET_PC, 0xa5, 5);

	/* FIXME: check ack from card, but it works... */

	pcipcwd_private.card_status = DISABLED;
	return rc;

	return 0;
}

static void
pcipcwd_get_relay2_mode(void)
{
	if (inb_p(pcipcwd_private.io_addr + 1) & WD_PCI_R2DS)
		pcipcwd_private.temp_mode_relay2 = NO_TEMPERATURE_TRIP;
	else
		pcipcwd_private.temp_mode_relay2 = TEMPERATURE_TRIP;

	if (debug >= DEBUG)
		printk(KERN_INFO PFX "mode of relay 2 is: %s\n",
		       (pcipcwd_private.temp_mode_relay2 ==
			TEMPERATURE_TRIP) ? "temp" : "notemp");
}

static int
pcipcwd_set_relay(char *sub_command)
{
	char *new_sub_command = sub_command;
	int i;

	while (*sub_command == ' ')
		sub_command++;
	if (sub_command == new_sub_command) {
		printk("relay value: %d\n", send_command_ret8(CMD_READ_RELAY_CONTROL, 0, 0));
		return 0;
	}

	i = send_command_ret8(CMD_READ_RELAY_CONTROL, 0, 0);	// read the relays

	if (strcmp(sub_command, "1on") == 0) {
		i = send_command_ret8(CMD_WRITE_RELAY_CONTROL, 0, (i | 1) & (0xFF - 4));
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "relay 1on\n");
	} else if (strcmp(sub_command, "1off") == 0) {
		i = send_command_ret8(CMD_WRITE_RELAY_CONTROL, 0, (i | 4) & (0xFF - 1));
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "relay 1off\n");

	} else if (strcmp(sub_command, "2on") == 0) {
		i = send_command_ret8(CMD_WRITE_RELAY_CONTROL, 0, (i | 2) & (0xFF - 8));
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "relay 2on\n");

	} else if (strcmp(sub_command, "2off") == 0) {
		i = send_command_ret8(CMD_WRITE_RELAY_CONTROL, 0, (i | 8) & (0xFF - 2));
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "relay 2off\n");

	} else if (strcmp(sub_command, "1invert") == 0) {
		if (debug >= DEBUG) {
			printk(KERN_DEBUG PFX "before: 0x%02x\n", i);
			printk(KERN_DEBUG PFX "sending: 0x%02x\n", i | 0x10);
		}
		i = send_command_ret8(CMD_WRITE_RELAY_CONTROL, 0, i | 0x10);
		if (debug >= DEBUG) {
			i = send_command_ret8(CMD_READ_RELAY_CONTROL, 0, 0);	// read the relays
			printk(KERN_DEBUG PFX "after: 0x%02x\n", i);
		}
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "relay 1invert\n");

	} else if (strcmp(sub_command, "1noinvert") == 0) {
		if (debug >= DEBUG) {
			printk(KERN_DEBUG PFX "before: 0x%02x\n", i);
			printk(KERN_DEBUG PFX "sending: 0x%02x\n",
			       i & (0xFF - 0x10));
		}
		i = send_command_ret8(CMD_WRITE_RELAY_CONTROL, 0, i & (0xFF - 0x10));
		if (debug >= DEBUG) {
			i = send_command_ret8(CMD_READ_RELAY_CONTROL, 0, 0);	// read the relays
			printk(KERN_DEBUG PFX "after: 0x%2x\n", i);
		}
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "relay 1noinvert\n");

	} else if (strcmp(sub_command, "1nvinvert") == 0) {
		if (debug >= DEBUG) {
			printk(KERN_DEBUG PFX "before: 0x%02x\n", i);
			printk(KERN_DEBUG PFX "sending: 0x%02x\n", i | 0x80);
		}
		i = send_command_ret8(CMD_WRITE_RELAY_CONTROL, 0, i | 0x80);
		if (debug >= DEBUG) {
			i = send_command_ret8(CMD_READ_RELAY_CONTROL, 0, 0);	// read the relays
			printk(KERN_DEBUG PFX "after: 0x%02x\n", i);
		}
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "relay 1nvinvert\n");

	} else if (strcmp(sub_command, "1nvnoinvert") == 0) {
		if (debug >= DEBUG) {
			printk(KERN_DEBUG PFX "before: 0x%02x\n", i);
			printk(KERN_DEBUG PFX "sending: 0x%02x\n",
			       i & (0xFF - 0x80));
		}
		i = send_command_ret8(CMD_WRITE_RELAY_CONTROL, 0, i & (0xFF - 0x80));
		if (debug >= DEBUG) {
			i = send_command_ret8(CMD_READ_RELAY_CONTROL, 0, 0);	// read the relays
			printk(KERN_DEBUG PFX "after: 0x%02x\n", i);
		}
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "relay 1nvnoinvert\n");

	} else if (strcmp(sub_command, "2notemp") == 0) {
		outb_p(WD_PCI_R2DS, pcipcwd_private.io_addr + 1);
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "relay 2notemp\n");
		pcipcwd_get_relay2_mode();

	} else if (strcmp(sub_command, "2temp") == 0) {
		outb_p(0x00, pcipcwd_private.io_addr + 1);
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "relay 2temp\n");
		pcipcwd_get_relay2_mode();

	} else {
		printk(KERN_ERR PFX
		       "illegal relay argument '%s', should be 1on, 1off, 2on, 2off, 1invert, 1noinvert, 1nvinvert, 1nvnoinvert, 2temp or 2notemp\n",
		       sub_command);
		return 1;
	}
	return 0;
}

static int
pcipcwd_set_dio(char *sub_command)
{
	char *new_sub_command = sub_command;
	int i;

	while (*sub_command == ' ')
		sub_command++;
	if (sub_command == new_sub_command) {
		printk("dio value: %d\n", (inb(pcipcwd_private.io_addr + 7) & 0xF0) >> 4);
		return 0;
	}
	i = simple_strtoul(sub_command, NULL, 0);
	if ((i >= 0) && (i < 16)) {
		outb_p(i, pcipcwd_private.io_addr + 7);
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "dio %d\n", i);
	} else {
		printk(KERN_ERR PFX
		       "illegal dio argument '%s', should be a number between 0 and 15\n",
		       sub_command);
		return 1;
	}
	return 0;
}

static int
pcipcwd_set_nv_timeout_pci(char *sub_command)
{
	char *new_sub_command = sub_command;
	int i;

	while (*sub_command == ' ')
		sub_command++;
	if (sub_command == new_sub_command) {
		printk("nvtimeout value: %d\n", send_command_ret16(CMD_READ_NV_WATCHDOG_TIMEOUT, 0, 0));
		return 0;
	}
	i = simple_strtoul(sub_command, NULL, 0);
	if (i == 0)
		printk("returning to default DIP switch settings.\n");

	if ((i >= 0) && (i <= 0xFFFF)) {
		send_command_ret16(CMD_WRITE_NV_WATCHDOG_TIMEOUT, i / 256, i % 256);
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "nvtimeout %d\n", i);
	} else {
		printk(KERN_ERR PFX
		       "illegal nvtimeout argument '%s', should be a number between 0 and 65535\n",
		       sub_command);
		return 1;
	}
	return 0;
}

static int
pcipcwd_set_timeout_pci(char *sub_command)
{
	char *new_sub_command = sub_command;
	int i;

	while (*sub_command == ' ')
		sub_command++;
	if (sub_command == new_sub_command) {
		printk("timeout value: %d\n", send_command_ret16(CMD_READ_WATCHDOG_TIMEOUT, 0, 0));
		return 0;
	}
	i = simple_strtoul(sub_command, NULL, 0);
	if (!pcipcwd_set_heartbeat(i)) {
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "timeout %d\n", heartbeat);
	} else {
		printk(KERN_ERR PFX
		       "illegal timeout argument '%s', should be a number between 0 and 65535\n",
		       sub_command);
		return 1;
	}
	return 0;
}

static int
pcipcwd_set_nv_arm_pci(char *sub_command)
{
	char *new_sub_command = sub_command;
	int i;

	while (*sub_command == ' ')
		sub_command++;
	if (sub_command == new_sub_command) {
		printk("nvarm value: %d\n", send_command_ret16(CMD_READ_NV_ARM_TIME, 0, 0));
		return 0;
	}
	i = simple_strtoul(sub_command, NULL, 0);
	if ((i >= 0) && (i <= 0xFFFF)) {
		send_command_ret16(CMD_WRITE_NV_ARM_TIME, i / 256, i % 256);
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "nvarm %d\n", i);
	} else {
		printk(KERN_ERR PFX
		       "illegal nvarm argument '%s', should be a number between 0 and 65535\n",
		       sub_command);
		return 1;
	}
	return 0;
}

static int
pcipcwd_set_arm_pci(char *sub_command)
{
	char *new_sub_command = sub_command;
	int i;

	while (*sub_command == ' ')
		sub_command++;
	if (sub_command == new_sub_command) {
		printk("arm value: %d\n", send_command_ret16(CMD_READ_ARM_TIME, 0, 0));
		return 0;
	}
	i = simple_strtoul(sub_command, NULL, 0);
	if ((i >= 0) && (i <= 0xFFFF)) {
		if (send_command_ret8(CMD_WRITE_ARM_TIME, i / 256, i % 256) == 0) {
			printk(KERN_INFO PFX
			       "arm failed, board already armed!\n");
			return 1;
		}
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "arm %d\n", i);
	} else {
		printk(KERN_ERR PFX
		       "illegal arm argument '%s', should be a number between 0 and 65535\n",
		       sub_command);
		return 1;
	}
	return 0;
}

static int
pcipcwd_set_tempoffset_pci(char *sub_command)
{
	char *new_sub_command = sub_command;
	int i;

	while (*sub_command == ' ')
		sub_command++;
	if (sub_command == new_sub_command) {
		printk("tempoffset value: %d\n", (inb(pcipcwd_private.io_addr + 2) & 0x0F));
		return 0;
	}

	i = simple_strtoul(sub_command, NULL, 0);
	if ((i >= 0) && (i < 16)) {
		outb_p(i, pcipcwd_private.io_addr + 2);
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "tempoffset %d\n", i);
	} else {
		printk(KERN_ERR PFX
		       "illegal tempoffset argument '%s', should be a number between 0 and 15\n",
		       sub_command);
		return 1;
	}
	return 0;
}

static void
pcipcwd_user_help(void)
{
	printk(" pcwd proc command interface help:\n");
	printk("   ping, pong, tickle - tickle the timer\n");
	printk("   hardreset - reset pc\n");
	printk("   enable, start - start the watchdog\n");
	printk("   disable, stop - stop the watchdog\n");
	printk("   clear - clear trip status and led\n");
	printk("   verbose, debug, quiet - debuging level\n");
	printk("   fahrenheit, celsius - display temp in\n");
	printk(" extended commands:\n");
	printk("   arm [val]\n");
	printk("   nvarm [val]\n");
	printk("   timeout [val]\n");
	printk("   nvtimeout [val]\n");
	printk("   dio [val]\n");
	printk("   relay [1on,1off,2on,2off,1invert,1noinvert,\n");
	printk("          1nvinvert,1nvnoinvert,2temp,2notemp]\n");
	printk("   tempoffset [val]\n");
}

static int
pcipcwd_user_command(char *user_command)
{
	if ((strcmp(user_command, "ping") == 0)
	    || (strcmp(user_command, "pong") == 0)
	    || (strcmp(user_command, "tickle") == 0)) {
		pcipcwd_keepalive();
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "ping...\n");
	} else if ((strcmp(user_command, "enable") == 0)
	    || (strcmp(user_command, "start") == 0)) {
		if (!pcipcwd_start()) {
			if (debug >= VERBOSE)
				printk(KERN_INFO PFX "watchdog enabled.\n");
		}
	} else if ((strcmp(user_command, "disable") == 0)
	    || (strcmp(user_command, "stop") == 0)) {
		if (!pcipcwd_stop()) {
			if (debug >= VERBOSE)
				printk(KERN_INFO PFX "watchdog disabled!\n");
		}
	} else if (strcmp(user_command, "clear") == 0) {
		pcipcwd_clear_status();
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "cleared trip status\n");
	} else if (strcmp(user_command, "hardreset") == 0) {
		pcipcwd_reset_pc();
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "hardreset PC!\n");
	} else if (strcmp(user_command, "quiet") == 0) {
		debug = QUIET;
	} else if (strcmp(user_command, "verbose") == 0) {
		debug = VERBOSE;
		printk(KERN_INFO PFX "verbose mode on.\n");
	} else if (strcmp(user_command, "debug") == 0) {
		debug = DEBUG;
		printk(KERN_INFO PFX "debug mode on.\n");
	} else if (strcmp(user_command, "celsius") == 0) {
		proc_temp_mode = CELSIUS;
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "display temp in Celsius.\n");
	} else if (strcmp(user_command, "fahrenheit") == 0) {
		proc_temp_mode = FAHRENHEIT;
		if (debug >= VERBOSE)
			printk(KERN_INFO PFX "display temp in Fahrenheit.\n");
	} else if (strcmp(user_command, "help") == 0) {
		pcipcwd_user_help();
	} else if (strncmp(user_command, "arm", 3) == 0) {
		if (pcipcwd_set_arm_pci(&user_command[3])) {
			printk(KERN_ERR PFX "arm FAILED.\n");
			return 0;
		}
	} else if (strncmp(user_command, "nvarm", 5) == 0) {
		if (pcipcwd_set_nv_arm_pci(&user_command[5])) {
			printk(KERN_ERR PFX "nvarm FAILED.\n");
			return 0;
		}
	} else if (strncmp(user_command, "timeout", 7) == 0) {
		if (pcipcwd_set_timeout_pci(&user_command[7])) {
			printk(KERN_ERR PFX "timeout FAILED.\n");
			return 0;
		}
	} else if (strncmp(user_command, "nvtimeout", 9) == 0) {
		if (pcipcwd_set_nv_timeout_pci(&user_command[9])) {
			printk(KERN_ERR PFX "nvtimeout FAILED.\n");
			return 0;
		}
	} else if (strncmp(user_command, "dio", 3) == 0) {
		if (pcipcwd_set_dio(&user_command[3])) {
			printk(KERN_ERR PFX "dio FAILED.\n");
			return 0;
		}
	} else if (strncmp(user_command, "relay", 5) == 0) {
		if (pcipcwd_set_relay(&user_command[5])) {
			printk(KERN_ERR PFX "relay FAILED.\n");
			return 0;
		}
	} else if (strncmp(user_command, "tempoffset", 10) == 0) {
		if (pcipcwd_set_tempoffset_pci(&user_command[10])) {
			printk(KERN_ERR PFX "tempoffset FAILED.\n");
			return 0;
		}
	} else {
		printk(KERN_ERR "illegal user command: '%s'\n", user_command);
		pcipcwd_user_help();
		return 0;
	}
	return 1;
}

static int
pcipcwd_write_proc(struct file *file, const char *buffer,
		unsigned long count, void *data)
{
	char command_buffer[80];
	int rc, length;

	/* write only allowed for root users */
	if (current->euid != 0)
		return -EACCES;

	if (count > sizeof (command_buffer) - 1)
		return -EINVAL;

	if (copy_from_user(command_buffer, buffer, count))
		return -EFAULT;

	command_buffer[count] = '\0';
	length = strlen(command_buffer);
	if (command_buffer[length - 1] == '\n')
		command_buffer[--length] = '\0';

	spin_lock(&pcipcwd_private.proc_lock);
	rc = pcipcwd_user_command(command_buffer);
	spin_unlock(&pcipcwd_private.proc_lock);
	return (rc ? count : -EBUSY);
}

/* This macro frees the machine specific function from bounds checking and
 *  * this like that... [from nvram.c]*/
#define PRINT_PROC(fmt,args...)				\
    do {						\
	*len += sprintf( buffer+*len, fmt, ##args );	\
	if (*begin + *len > offset + size)		\
	    return( 0 );				\
	if (*begin + *len < offset) {			\
	    *begin += *len;				\
	    *len = 0;					\
	}						\
    } while(0)

static int
pcipcwd_proc_info(unsigned char *buf, char *buffer, int *len,
	       off_t * begin, off_t offset, int size)
{
	int sw, i, j;

	PRINT_PROC("PCI-PC Watchdog Driver %s\n", pcwd_driver_version);
	PRINT_PROC("Firmware     : Version %s\n", pcipcwd_private.fw_ver_str);
	PRINT_PROC("Option Switch: Delay Time ");
	sw = pcipcwd_get_option_switches();
	switch (sw & 0x07) {
	case 0:
		PRINT_PROC("5 s");
		break;
	case 1:
		PRINT_PROC("10 s");
		break;
	case 2:
		PRINT_PROC("30 s");
		break;
	case 3:
		PRINT_PROC("1 min");
		break;
	case 4:
		PRINT_PROC("5 min");
		break;
	case 5:
		PRINT_PROC("10 min");
		break;
	case 6:
		PRINT_PROC("30 min");
		break;
	case 7:
		PRINT_PROC("1 h");
		break;
	}
	PRINT_PROC((sw & 0x08) ? ", POD on" : ", POD off");
	PRINT_PROC((sw & 0x10) ? ", TRE on" : ", TRE off");
	PRINT_PROC((sw & 0x20) ? ", R2M on" : ", R2M off");
	PRINT_PROC((sw & 0x40) ? ", R1M on" : ", R1M off");
	PRINT_PROC((sw & 0x80) ? ", RTM on\n" : ", RTM off\n");

	sw = send_command_ret8(CMD_READ_RELAY_CONTROL, 0, 0);
	PRINT_PROC("Relays       : ");
	PRINT_PROC((sw & 0x01) ? "Relay 1 on" : "Relay 1 off");
	PRINT_PROC((sw & 0x02) ? ", Relay 2 on\n" : ", Relay 2 off\n");
	PRINT_PROC("               ");
	PRINT_PROC((sw & 0x10) ? "Relay 1 invert, " : "");
	PRINT_PROC((sw & 0x80) ? "Relay 1 invert in nvram, " : "");
	PRINT_PROC(pcipcwd_private.temp_mode_relay2 ? "Relay 2 temp\n" : "Relay 2 notemp\n");

	i = send_command_ret16(CMD_READ_NV_ARM_TIME, 0, 0);	// ARM NVRAM
	j = send_command_ret16(CMD_READ_ARM_TIME, 0, 0);	// ARM current
	PRINT_PROC("Arm Time     : %d s (NVRAM: %d s)\n", j, i);

	i = send_command_ret16(CMD_READ_NV_WATCHDOG_TIMEOUT, 0, 0);	// Time NVRAM
	j = send_command_ret16(CMD_READ_WATCHDOG_TIMEOUT, 0, 0);	// Time current
	PRINT_PROC("Timer        : %d s (NVRAM: %d s)\n", j, i);

	sw = send_command_ret8(CMD_GET_STATUS, 0, 0);
	PRINT_PROC("Status       : ");
	PRINT_PROC((sw & 0x01) ? "NVRAM OK" : "NVRAM FAILED");
	PRINT_PROC((sw & 0x02) ? ", TEMP SENSOR OK" : ", TEMP SENSOR FAILED");
	PRINT_PROC((sw & 0x80) ? ", POD COMPLETED.\n" : ", POD ACTIVE.\n");

	sw = inb(pcipcwd_private.io_addr + 7);
	PRINT_PROC("Digi I/O Port: OUT=%d, IN=%d\n", (sw & 0x0F),
		   ((sw & 0xF0) >> 4));

	sw = send_command_ret8(CMD_GET_CLEAR_RESET_COUNT, 0, 0);
	PRINT_PROC("Reset Count  : %d\n", sw);

	sw = inb(pcipcwd_private.io_addr + 2) & 0x0F;
	PRINT_PROC("Temp.TripPt. : LOWER=%d C UPPER=%d C\n", 46 + sw,
		   56 + sw);

	PRINT_PROC("Temperature  : ");
	if (pcipcwd_private.supports_temp && (!pcipcwd_get_temperature(&i, proc_temp_mode))) {
		PRINT_PROC("%d ", i);
		PRINT_PROC((proc_temp_mode == CELSIUS) ? "C\n" :"F\n");
	} else {
		PRINT_PROC("Card without temp option\n");
	}

	if (pcipcwd_private.card_status == ENABLED)
		PRINT_PROC("Card status  : Timer enabled\n");
	else
		PRINT_PROC("Card status  : Timer disabled\n");

	sw = inb_p(pcipcwd_private.io_addr + 1) & WD_PCI_WTRP;
	PRINT_PROC("Reboot status: ");
	PRINT_PROC(sw ? "Tripped\n" : "Not tripped\n");

	if ((pcipcwd_private.boot_status & WDIOF_CARDRESET) && (pcipcwd_private.boot_status & WDIOF_OVERHEAT)) {
		PRINT_PROC("Boot status  : CPU Overheat  + Watchdog caused reboot\n");
	} else if (pcipcwd_private.boot_status & WDIOF_CARDRESET) {
		PRINT_PROC("Boot status  : Watchdog caused reboot\n");
	} else if (pcipcwd_private.boot_status & WDIOF_OVERHEAT) {
		PRINT_PROC("Boot status  : CPU Overheat\n");
	} else {
		PRINT_PROC("Boot status  : Cold boot or reset\n");
	}

	// PRINT_PROC("Ping Count   : %d\n", pcipcwd_private.ping_counter);
	return 1;
}

static int
pcipcwd_read_proc(char *buffer, char **start, off_t offset,
	       int size, int *eof, void *data)
{
	int len = 0;
	off_t begin = 0;

	spin_lock(&pcipcwd_private.proc_lock);
	*eof = pcipcwd_proc_info(NULL, buffer, &len, &begin, offset, size);
	spin_unlock(&pcipcwd_private.proc_lock);

	if (offset >= begin + len)
		return (0);
	*start = buffer + (offset - begin);	// CORRECTED !!!!
	return (size < begin + len - offset ? size : begin + len - offset);
}
#endif				/* PROC_FS */

/*
 *	/dev/watchdog handling
 */

static ssize_t pcipcwd_write(struct file *file, const char __user *data,
			     size_t len, loff_t *ppos)
{
	/* See if we got the magic character 'V' and reload the timer */
	if (len) {
		if (!nowayout) {
			size_t i;

			/* note: just in case someone wrote the magic character
			 * five months ago... */
			expect_release = 0;

			/* scan to see whether or not we got the magic character */
			for (i = 0; i != len; i++) {
				char c;
				if(get_user(c, data+i))
					return -EFAULT;
				if (c == 'V')
					expect_release = 42;
			}
		}

		/* someone wrote to us, we should reload the timer */
		pcipcwd_keepalive();
	}
	return len;
}

static int pcipcwd_ioctl(struct inode *inode, struct file *file,
			  unsigned int cmd, unsigned long arg)
{
	void __user *argp = (void __user *)arg;
	int __user *p = argp;
	static struct watchdog_info ident = {
		.options =		WDIOF_OVERHEAT |
					WDIOF_CARDRESET |
					WDIOF_KEEPALIVEPING |
					WDIOF_SETTIMEOUT |
					WDIOF_MAGICCLOSE,
		.firmware_version =	1,
		.identity =		WATCHDOG_DRIVER_NAME,
	};

	switch (cmd) {
		case WDIOC_GETSUPPORT:
			return copy_to_user(argp, &ident,
				sizeof (ident)) ? -EFAULT : 0;

		case WDIOC_GETSTATUS:
		{
			int status;

			pcipcwd_get_status(&status);

			return put_user(status, p);
		}

		case WDIOC_GETBOOTSTATUS:
			return put_user(pcipcwd_private.boot_status, p);

		case WDIOC_GETTEMP:
		{
			int temperature;

			if (pcipcwd_get_temperature(&temperature, FAHRENHEIT))
				return -EFAULT;

			return put_user(temperature, p);
		}

		case WDIOC_KEEPALIVE:
			pcipcwd_keepalive();
			return 0;

		case WDIOC_SETOPTIONS:
		{
			int new_options, retval = -EINVAL;

			if (get_user (new_options, p))
				return -EFAULT;

			if (new_options & WDIOS_DISABLECARD) {
				if (pcipcwd_stop())
					return -EIO;
				retval = 0;
			}

			if (new_options & WDIOS_ENABLECARD) {
				if (pcipcwd_start())
					return -EIO;
				retval = 0;
			}

			if (new_options & WDIOS_TEMPPANIC) {
				temp_panic = 1;
				retval = 0;
			}

			return retval;
		}

		case WDIOC_SETTIMEOUT:
		{
			int new_heartbeat;

			if (get_user(new_heartbeat, p))
				return -EFAULT;

			if (pcipcwd_set_heartbeat(new_heartbeat))
			    return -EINVAL;

			pcipcwd_keepalive();
			/* Fall */
		}

		case WDIOC_GETTIMEOUT:
			return put_user(heartbeat, p);

		default:
			return -ENOIOCTLCMD;
	}
}

static int pcipcwd_open(struct inode *inode, struct file *file)
{
	/* /dev/watchdog can only be opened once */
	if (test_and_set_bit(0, &is_active)) {
		if (debug >= VERBOSE)
			printk(KERN_ERR PFX "Attempt to open already opened device.\n");
		return -EBUSY;
	}

	/* Activate */
	pcipcwd_start();
	pcipcwd_keepalive();
	return nonseekable_open(inode, file);
}

static int pcipcwd_release(struct inode *inode, struct file *file)
{
	/*
	 *      Shut off the timer.
	 */
	if (expect_release == 42) {
		pcipcwd_stop();
	} else {
		printk(KERN_CRIT PFX "Unexpected close, not stopping watchdog!\n");
		pcipcwd_keepalive();
	}
	expect_release = 0;
	clear_bit(0, &is_active);
	return 0;
}

/*
 *	/dev/temperature handling
 */

static ssize_t pcipcwd_temp_read(struct file *file, char __user *data,
				size_t len, loff_t *ppos)
{
	int temperature;

	if (pcipcwd_get_temperature(&temperature, FAHRENHEIT))
		return -EFAULT;

	if (copy_to_user (data, &temperature, 1))
		return -EFAULT;

	return 1;
}

static int pcipcwd_temp_open(struct inode *inode, struct file *file)
{
	if (!pcipcwd_private.supports_temp)
		return -ENODEV;

	return nonseekable_open(inode, file);
}

static int pcipcwd_temp_release(struct inode *inode, struct file *file)
{
	return 0;
}

/*
 *	Notify system
 */

static int pcipcwd_notify_sys(struct notifier_block *this, unsigned long code, void *unused)
{
	if (code==SYS_DOWN || code==SYS_HALT) {
		/* Turn the WDT off */
		pcipcwd_stop();
		printk(KERN_INFO PFX "reboot notifier has disabled the card!\n");
	}

	return NOTIFY_DONE;
}

/*
 *	Kernel Interfaces
 */

static struct file_operations pcipcwd_fops = {
	.owner =	THIS_MODULE,
	.llseek =	no_llseek,
	.write =	pcipcwd_write,
	.ioctl =	pcipcwd_ioctl,
	.open =		pcipcwd_open,
	.release =	pcipcwd_release,
};

static struct miscdevice pcipcwd_miscdev = {
	.minor =	WATCHDOG_MINOR,
	.name =		"watchdog",
	.fops =		&pcipcwd_fops,
};

static struct file_operations pcipcwd_temp_fops = {
	.owner =	THIS_MODULE,
	.llseek =	no_llseek,
	.read =		pcipcwd_temp_read,
	.open =		pcipcwd_temp_open,
	.release =	pcipcwd_temp_release,
};

static struct miscdevice pcipcwd_temp_miscdev = {
	.minor =	TEMP_MINOR,
	.name =		"temperature",
	.fops =		&pcipcwd_temp_fops,
};

static struct notifier_block pcipcwd_notifier = {
	.notifier_call =	pcipcwd_notify_sys,
};

/*
 *	Init & exit routines
 */

static int __devinit pcipcwd_card_init(struct pci_dev *pdev,
		const struct pci_device_id *ent)
{
	int ret = -EIO;

	cards_found++;
	if (cards_found == 1)
		printk(KERN_INFO PFX "%s by <holger@eiboeck.de>\n",
			pcwd_driver_version);

	if (cards_found > 1) {
		printk(KERN_ERR PFX "This driver only supports 1 device\n");
		return -ENODEV;
	}

	if (pci_enable_device(pdev)) {
		printk(KERN_ERR PFX "Not possible to enable PCI Device\n");
		return -ENODEV;
	}

	if (pci_resource_start(pdev, 0) == 0x0000) {
		printk(KERN_ERR PFX "No I/O-Address for card detected\n");
		ret = -ENODEV;
		goto err_out_disable_device;
	}

	pcipcwd_private.io_addr = pci_resource_start(pdev, 0);

	if (pci_request_regions(pdev, WATCHDOG_NAME)) {
		printk(KERN_ERR PFX "I/O address 0x%04x already in use\n",
			(int) pcipcwd_private.io_addr);
		ret = -EIO;
		goto err_out_disable_device;
	}

	/* get the boot_status */
	pcipcwd_get_status(&pcipcwd_private.boot_status);

	/* clear the "card caused reboot" flag */
	pcipcwd_clear_status();

	/* disable card */
	pcipcwd_stop();

	/* Check whether or not the card supports the temperature device */
	pcipcwd_check_temperature_support();

	/* Show info about the card itself */
	pcipcwd_show_card_info();

	/* Check that the heartbeat value is within it's range ; if not reset to the default */
	if (pcipcwd_set_heartbeat(heartbeat)) {
		pcipcwd_set_heartbeat(WATCHDOG_HEARTBEAT);
		printk(KERN_INFO PFX "heartbeat value must be 0<heartbeat<65536, using %d\n",
			WATCHDOG_HEARTBEAT);
	}

	ret = register_reboot_notifier(&pcipcwd_notifier);
	if (ret != 0) {
		printk(KERN_ERR PFX "cannot register reboot notifier (err=%d)\n",
			ret);
		goto err_out_release_region;
	}

	if (pcipcwd_private.supports_temp) {
		ret = misc_register(&pcipcwd_temp_miscdev);
		if (ret != 0) {
			printk(KERN_ERR PFX "cannot register miscdev on minor=%d (err=%d)\n",
				TEMP_MINOR, ret);
			goto err_out_unregister_reboot;
		}
	}

	ret = misc_register(&pcipcwd_miscdev);
	if (ret != 0) {
		printk(KERN_ERR PFX "cannot register miscdev on minor=%d (err=%d)\n",
			WATCHDOG_MINOR, ret);
		goto err_out_misc_deregister;
	}

#ifdef CONFIG_PROC_FS
	if ((pcipcwd_private.proc_pcwd =
	     create_proc_entry(WATCHDOG_NAME, S_IFREG | S_IRUGO | S_IWUSR, 0))) {
		pcipcwd_private.proc_pcwd->read_proc = pcipcwd_read_proc;
		pcipcwd_private.proc_pcwd->write_proc = pcipcwd_write_proc;
	}
#endif

	if (nowayout)
		printk(KERN_INFO PFX "Watchdog cannot be stopped once started\n");

	printk(KERN_INFO PFX "initialized. heartbeat=%d sec (nowayout=%d)\n",
		heartbeat, nowayout);

	return 0;

err_out_misc_deregister:
	if (pcipcwd_private.supports_temp)
		misc_deregister(&pcipcwd_temp_miscdev);
err_out_unregister_reboot:
	unregister_reboot_notifier(&pcipcwd_notifier);
err_out_release_region:
	pci_release_regions(pdev);
err_out_disable_device:
	pci_disable_device(pdev);
	return ret;
}

static void __devexit pcipcwd_card_exit(struct pci_dev *pdev)
{
#ifdef CONFIG_PROC_FS
	if (pcipcwd_private.proc_pcwd)
		remove_proc_entry(WATCHDOG_NAME, 0);
#endif

	/* Stop the timer before we leave */
	if (!nowayout)
		pcipcwd_stop();

	/* Deregister */
	misc_deregister(&pcipcwd_miscdev);
	if (pcipcwd_private.supports_temp)
		misc_deregister(&pcipcwd_temp_miscdev);
	unregister_reboot_notifier(&pcipcwd_notifier);
	pci_release_regions(pdev);
	pci_disable_device(pdev);
	cards_found--;
}

static struct pci_device_id pcipcwd_pci_tbl[] = {
	{ PCI_VENDOR_ID_QUICKLOGIC, PCI_DEVICE_ID_WATCHDOG_PCIPCWD,
		PCI_ANY_ID, PCI_ANY_ID, },
	{ 0 },			/* End of list */
};
MODULE_DEVICE_TABLE(pci, pcipcwd_pci_tbl);

static struct pci_driver pcipcwd_driver = {
	.name		= WATCHDOG_NAME,
	.id_table	= pcipcwd_pci_tbl,
	.probe		= pcipcwd_card_init,
	.remove		= __devexit_p(pcipcwd_card_exit),
};

static int __init pcipcwd_init_module(void)
{
	spin_lock_init(&pcipcwd_private.io_lock);
#ifdef CONFIG_PROC_FS
	spin_lock_init(&pcipcwd_private.proc_lock);
#endif

	pcipcwd_get_version();
	if (proc_temp_mode)
		proc_temp_mode = FAHRENHEIT;

	return pci_register_driver(&pcipcwd_driver);
}

static void __exit pcipcwd_cleanup_module(void)
{
	pci_unregister_driver(&pcipcwd_driver);

	printk(KERN_INFO PFX "Watchdog Module Unloaded.\n");
}

module_init(pcipcwd_init_module);
module_exit(pcipcwd_cleanup_module);

MODULE_AUTHOR("Holger Eiboeck & Wim Van Sebroeck");
MODULE_DESCRIPTION("Driver for the Berkshire PCI-PC watchdog cards");
MODULE_LICENSE("GPL");
MODULE_ALIAS_MISCDEV(WATCHDOG_MINOR);
MODULE_ALIAS_MISCDEV(TEMP_MINOR);

